import dataclasses
import enum
import inspect
from abc import ABC, abstractmethod
from datetime import datetime, timezone
from functools import partial
from types import UnionType
from typing import (
    Annotated,
    Any,
    Generic,
    Literal,
    Optional,
    Protocol,
    Sequence,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
    runtime_checkable,
)
from uuid import UUID

import msgspec
import numpy as np
from psycopg.types.json import Jsonb
from uuid_utils.compat import uuid7

from vechord.typing import Self

__all__ = [
    "DefaultDocument",
    "ForeignKey",
    "Keyword",
    "KeywordIndex",
    "MultiVectorIndex",
    "PrimaryKeyAutoIncrease",
    "PrimaryKeyUUID",
    "Table",
    "Vector",
    "VectorIndex",
    "create_chunk_with_dim",
]


@runtime_checkable
class VechordType(Protocol):
    @classmethod
    def schema(cls) -> str:
        pass

    @classmethod
    def psql_type(cls) -> str:
        pass


class VectorMeta(type):
    def __getitem__(self, dim: int):
        if not isinstance(dim, int) or dim <= 0:
            raise ValueError(
                f"dim must be a positive integer, not `{type(dim)}({dim})`"
            )
        return create_vector_type(dim)


V = TypeVar("V")


class Vector(Generic[V], metaclass=VectorMeta):
    """Vector type with fixed dimension.

    User can assign `np.ndarray` with `np.float32` type or `list[float]` type.
    """

    def __init__(self, *args, **kwargs):
        if self.__class__ is Vector:
            raise ValueError("Use Vector[dim] to create a vector type")

    @classmethod
    def schema(cls) -> str:
        raise NotImplementedError("Should be implemented by the subclass: Vector[dim]")


def create_vector_type(dim: int) -> Type[Vector]:
    name = f"Vector[{dim}]"

    class SpecificVector(Vector, np.ndarray):
        nonlocal dim
        _dim: int = dim

        def __new__(cls, vec: list[float] | np.ndarray):
            if isinstance(vec, np.ndarray):
                if vec.shape != (dim,):
                    raise ValueError(f"expected shape ({dim},), got {vec.shape}")
            elif isinstance(vec, list):
                if len(vec) != dim:
                    raise ValueError(f"expected length {dim}, got {len(vec)}")
                vec = np.array(vec, dtype=np.float32)
            else:
                raise ValueError("expected list or np.ndarray")
            return np.asarray(vec, dtype=np.float32).view(cls)

        @classmethod
        def schema(cls):
            return f"VECTOR({cls._dim})"

        @classmethod
        def psql_type(cls):
            return "vector"

        @classmethod
        def __json_schema__(cls) -> dict:
            return {
                "title": cls.__name__,
                "type": "array",
                "items": {"type": "number", "format": "float"},
                "minItems": cls._dim,
                "maxItems": cls._dim,
            }

        @classmethod
        def __json_encode__(cls, value: Self) -> list[float]:
            return value.tolist()

        @classmethod
        def __json_decode__(cls, value: list[float]) -> Self:
            return cls(value)

    SpecificVector.__name__ = name
    return SpecificVector


class ForeignKeyMeta(type):
    def __getitem__(self, ref):
        return create_foreign_key_type(ref)


F = TypeVar("F")


class ForeignKey(Generic[F], metaclass=ForeignKeyMeta):
    """Reference to another table's attribute as a foreign key.

    This should be used in the `Annotated[]` type hint.
    """

    def __init__(self, *args, **kwargs):
        raise NotImplementedError("Use ForeignKey[ref] to create a foreign key type")

    @classmethod
    def schema(cls) -> str:
        raise NotImplementedError(
            "Should be implemented by the subclass: ForeignKey[ref]"
        )


def create_foreign_key_type(ref) -> Type[ForeignKey]:
    name = f"ForeignKey[{ref}]"

    class SpecificForeignKey(ForeignKey):
        nonlocal ref
        _ref = ref

        def __init__(self, value):
            self.value = value

        @classmethod
        def schema(cls):
            ref_cls = cls._ref.__objclass__.__name__.lower()
            ref_val = cls._ref.__name__
            return f"REFERENCES {{namespace}}_{ref_cls}({ref_val}) ON DELETE CASCADE"

        @classmethod
        def psql_type(cls):
            return ""

    SpecificForeignKey.__name__ = name
    return SpecificForeignKey


class PrimaryKeyAutoIncrease(int):
    """Primary key with auto-increment ID type. (wrap ``int``)"""

    @classmethod
    def schema(cls) -> str:
        return "BIGINT GENERATED ALWAYS AS IDENTITY PRIMARY KEY"

    @classmethod
    def psql_type(cls) -> str:
        return "bigint"

    @classmethod
    def __json_schema__(cls) -> dict:
        return {
            "title": cls.__name__,
            "type": "integer",
            "format": "int64",
        }

    @classmethod
    def __json_encode__(cls, value: Self) -> int:
        return int(value)

    @classmethod
    def __json_decode__(cls, value: int) -> Self:
        if not isinstance(value, int):
            raise ValueError(f"expected int, got {type(value)}")
        return cls(value)


class PrimaryKeyUUID(UUID):
    """Primary key with UUID type. (wrap ``UUID``)

    This doesn't come with auto-generate, because PostgreSQL doesn't support UUID v7, while v4
    is purely random and not sortable.

    Choose this one over :class:`PrimaryKeyAutoIncrease` when you need universal uniqueness.

    We suggest to use:

    .. code-block:: python

        class MyTable(Table):
            uid: PrimaryKeyUUID = msgspec.field(default_factory=PrimaryKeyUUID.factory)
    """

    @classmethod
    def schema(cls) -> str:
        return "UUID PRIMARY KEY"

    @classmethod
    def psql_type(cls) -> str:
        return "uuid"

    @classmethod
    def factory(cls):
        return uuid7()

    @classmethod
    def __json_schema__(cls) -> dict:
        return {
            "title": cls.__name__,
            "type": "string",
            "format": "uuid",
        }

    @classmethod
    def __json_encode__(cls, value: Self) -> str:
        return str(value)

    @classmethod
    def __json_decode__(cls, value: str) -> Self:
        return cls(value)


class Keyword(str):
    """Keyword type for text search. (wrap ``str``)

    User can assign the `str` type, it will be tokenized and converted to
    `bm25vector` in PostgreSQL.
    """

    _model: Literal["bert_base_uncased", "wiki_tocken"] = "bert_base_uncased"

    @classmethod
    def schema(cls) -> str:
        return "bm25vector"

    @classmethod
    def psql_type(cls) -> str:
        return "bm25vector"

    @classmethod
    def with_model(cls, model: Literal["bert_base_uncased", "wiki_tocken"]) -> Type:
        cls._model = model
        return cls

    @classmethod
    def __json_schema__(cls) -> dict:
        return {
            "title": cls.__name__,
            "type": "string",
            "format": "keyword",
            "description": f"Keyword with model {cls._model}",
        }

    @classmethod
    def __json_encode__(cls, value: Self) -> str:
        return str(value)

    @classmethod
    def __json_decode__(cls, value: str) -> Self:
        return cls(value)


TYPE_TO_PSQL = {
    int: "BIGINT",
    str: "TEXT",
    bytes: "BYTEA",
    float: "FLOAT8",
    bool: "BOOLEAN",
    UUID: "UUID",
    datetime: "TIMESTAMPTZ",
    Jsonb: "JSONB",
}


def is_optional_type(typ: Type) -> bool:
    return (get_origin(typ) is Union or get_origin(typ) is UnionType) and type(
        None
    ) in get_args(typ)


def get_first_type_from_optional(typ: Type) -> Type:
    for arg in get_args(typ):
        if arg is not type(None):
            return arg
    raise ValueError(f"no non-None type in {typ}")


def get_first_arg(typ: Type) -> Type:
    """Get the first argument of a type, if it has one."""
    args = get_args(typ)
    if not args:
        raise ValueError(f"{typ} has no arguments")
    return args[0]


def is_list_of_vector_type(typ: Type) -> bool:
    return get_origin(typ) is list and issubclass(
        get_first_arg(typ).__class__, VectorMeta
    )


def py_type_to_psql_schema(typ) -> str:
    if is_optional_type(typ):
        typ = get_first_type_from_optional(typ)

    if get_origin(typ) is Annotated:
        origin, *meta = get_args(typ)
        schema = [py_type_to_psql_schema(origin)]
        for m in meta:
            if inspect.isclass(m) and issubclass(m, ForeignKey):
                schema.append(m.schema())
        return " ".join(schema)
    elif get_origin(typ) is list:
        return f"{py_type_to_psql_schema(get_first_arg(typ))}[]"
    if isinstance(typ, VechordType):
        return typ.schema()
    if typ in TYPE_TO_PSQL:
        return TYPE_TO_PSQL[typ]
    raise ValueError(f"unsupported type {typ}")


def py_type_to_psql_type(typ) -> str:
    if is_optional_type(typ):
        typ = get_first_type_from_optional(typ)

    if get_origin(typ) is Annotated:
        return py_type_to_psql_type(get_first_arg(typ))
    elif get_origin(typ) is list:
        return f"{py_type_to_psql_type(get_first_arg(typ))}[]"
    if isinstance(typ, VechordType):
        return typ.psql_type()
    if typ in TYPE_TO_PSQL:
        return TYPE_TO_PSQL[typ].lower()
    raise ValueError(f"unsupported type {typ}")


class VectorDistance(enum.Enum):
    L2 = "l2"
    COS = "cos"
    DOT = "dot"


class BaseIndex(ABC):
    name: str = ""
    index: str = ""
    op_name: str = ""
    op_symbol: str = ""

    def verify(self):
        """This will construct a new data-only struct to verify all the fields.

        This is to ensure that the `__post_init__` will not be called recursively.
        """
        fields = [
            (
                f.name,
                f.type,
                f.default
                if f.default is not f.default_factory
                else msgspec.field(default_factory=f.default_factory),
            )
            for f in dataclasses.fields(self)
        ]
        spec = msgspec.defstruct(f"Spec{self.__class__.__name__}", fields)
        instance = msgspec.convert(dataclasses.asdict(self), spec)
        for field in instance.__struct_fields__:
            setattr(self, field, getattr(instance, field))

    @abstractmethod
    def config(self) -> str:
        raise NotImplementedError


Idx = TypeVar("Idx", bound=BaseIndex)


class IndexColumn(msgspec.Struct, Generic[Idx], frozen=True, order=True):
    name: str
    index: Idx


@dataclasses.dataclass
class VectorIndex(BaseIndex):
    distance: VectorDistance = VectorDistance.L2
    lists: Optional[int] = None

    def __post_init__(self):
        self.verify()
        self.name = "vec_idx"
        self.index = "vchordrq"
        match self.distance:
            case VectorDistance.L2:
                self.op_symbol = "<->"
                self.op_name = "vector_l2_ops"
            case VectorDistance.COS:
                self.op_symbol = "<=>"
                self.op_name = "vector_cosine_ops"
            case VectorDistance.DOT:
                self.op_symbol = "<#>"
                self.op_name = "vector_ip_ops"

    def config(self):
        is_l2 = self.distance == VectorDistance.L2
        return f"""
residual_quantization = {"true" if is_l2 else "false"}
[build.internal]
lists = [{self.lists or ""}]
spherical_centroids = {"false" if is_l2 else "true"}
"""


@dataclasses.dataclass
class MultiVectorIndex(BaseIndex):
    lists: Optional[int] = None

    def __post_init__(self):
        self.verify()
        self.name = "multivec_idx"
        self.index = "vchordrq"
        self.op_name = "vector_maxsim_ops"
        self.op_symbol = "@#"

    def config(self):
        return f"""
residual_quantization = false
[build.internal]
lists = [{self.lists or ""}]
spherical_centroids = true
"""


@dataclasses.dataclass
class KeywordIndex(BaseIndex):
    model: Literal["bert_base_uncased", "wiki_tocken"] = "bert_base_uncased"

    def __post_init__(self):
        self.verify()
        self.name = "keyword_idx"
        self.index = "bm25"
        self.op_name = "bm25_ops"
        self.op_symbol = "<&>"

    def config(self):
        return ""


@dataclasses.dataclass
class UniqueIndex(BaseIndex):
    null_not_distinct: bool = False

    def __post_init__(self):
        self.verify()
        self.name = "unique_idx"

    def config(self):
        return "NULLS NOT DISTINCT" if self.null_not_distinct else ""


class Storage(msgspec.Struct):
    @classmethod
    def name(cls) -> str:
        return cls.__name__.lower()

    @classmethod
    def fields(cls) -> Sequence[str]:
        return cls.__struct_fields__

    @classmethod
    def partial_init(cls, **kwargs):
        fields = cls.fields()
        args = dict(zip(fields, [msgspec.UNSET] * len(fields), strict=False)) | kwargs
        return cls(**args)


E = TypeVar("E")


class AnyOf(msgspec.Struct, Generic[E]):
    """Select the records that match any of the given values."""

    values: Sequence[E]


class Table(Storage):
    """Base class for table definition."""

    @classmethod
    def table_schema(cls) -> Sequence[tuple[str, str]]:
        """Generate the table schema from the class attributes' type hints."""
        hints = get_type_hints(cls, include_extras=True)
        return tuple((name, py_type_to_psql_schema(typ)) for name, typ in hints.items())

    @classmethod
    def table_psql_types(cls) -> Sequence[tuple[str, str]]:
        """Generate the corresponding PostgreSQL types for each column."""
        hints = get_type_hints(cls, include_extras=True)
        return tuple((name, py_type_to_psql_type(typ)) for name, typ in hints.items())

    @classmethod
    def vector_column(cls) -> Optional[IndexColumn[VectorIndex]]:
        """Get the vector column name."""
        for name, typ in get_type_hints(cls, include_extras=True).items():
            if issubclass(typ.__class__, VectorMeta):
                return IndexColumn(name, VectorIndex())
            elif get_origin(typ) is Annotated and issubclass(
                get_first_arg(typ).__class__, VectorMeta
            ):
                for m in typ.__metadata__:
                    if isinstance(m, VectorIndex):
                        return IndexColumn(name, m)
        return None

    @classmethod
    def multivec_column(cls) -> Optional[IndexColumn[MultiVectorIndex]]:
        """Get the multivec column name."""
        for name, typ in get_type_hints(cls, include_extras=True).items():
            if is_list_of_vector_type(typ):
                return IndexColumn(name, MultiVectorIndex())
            elif get_origin(typ) is Annotated:
                inner_type = get_first_arg(typ)
                if is_list_of_vector_type(inner_type):
                    for m in typ.__metadata__:
                        if isinstance(m, MultiVectorIndex):
                            return IndexColumn(name, m)
        return None

    @classmethod
    def keyword_column(cls) -> Optional[IndexColumn[KeywordIndex]]:
        """Get the keyword column name."""
        for name, typ in get_type_hints(cls, include_extras=True).items():
            if typ is Keyword:
                return IndexColumn(name, KeywordIndex())
            elif get_origin(typ) is Annotated and get_first_arg(typ) is Keyword:
                for m in typ.__metadata__:
                    if isinstance(m, KeywordIndex):
                        return IndexColumn(name, m)
        return None

    @classmethod
    def unique_columns(cls) -> Sequence[IndexColumn]:
        """Get all the index columns."""
        columns = []
        for name, typ in get_type_hints(cls, include_extras=True).items():
            if get_origin(typ) is Annotated:
                for m in typ.__metadata__:
                    if isinstance(m, UniqueIndex):
                        columns.append(IndexColumn(name, m))
        return columns

    @classmethod
    def non_vec_columns(cls) -> Sequence[str]:
        """Get the column names that are not vector or keyword."""
        exclude = tuple(
            col.name if col else col
            for col in (
                cls.vector_column(),
                cls.keyword_column(),
                cls.multivec_column(),
            )
        )
        return tuple(field for field in cls.fields() if field not in exclude)

    @classmethod
    def keyword_tokenizer(cls) -> Optional[str]:
        """Get the keyword tokenizer."""
        for _, typ in get_type_hints(cls, include_extras=True).items():
            if typ is Keyword:
                return typ._model
            elif get_origin(typ) is Annotated and get_first_arg(typ) is Keyword:
                return get_first_arg(typ)._model
        return None

    @classmethod
    def primary_key(cls) -> Optional[str]:
        """Get the primary key column name."""
        for name, typ in get_type_hints(cls, include_extras=True).items():
            typ_cls = (
                get_first_type_from_optional(typ) if is_optional_type(typ) else typ
            )
            if inspect.isclass(typ_cls) and issubclass(
                typ_cls, (PrimaryKeyUUID, PrimaryKeyAutoIncrease)
            ):
                return name
        return None

    def todict(self) -> dict[str, Any]:
        """Convert the table instance to a dictionary.

        This will ignore the values like:

        - `msgspec.UNSET`
        - default value is None and the value is also None (mainly for PrimaryKeyAutoIncrease)
        """
        defaults = getattr(self, "__struct_defaults__", None)
        fields = self.fields()
        if not defaults:
            return {
                k: v
                for k, v in zip(fields, (getattr(self, f) for f in fields), strict=True)
                if v is not msgspec.UNSET
            }
        # ignore default values
        res = {}
        for k, d in zip(fields, defaults, strict=False):
            v = getattr(self, k)
            if v is not msgspec.UNSET and (d is msgspec.NODEFAULT or v is not d):
                res[k] = v
        return res


class DefaultDocument(Table, kw_only=True):
    """Default Document table class."""

    uid: PrimaryKeyUUID = msgspec.field(default_factory=PrimaryKeyUUID.factory)
    title: str = ""
    text: str
    created_at: datetime = msgspec.field(
        default_factory=partial(datetime.now, timezone.utc)
    )


class _DefaultChunk(Table, kw_only=True):
    """A placeholder for the chunk table class.

    This is not intended to be used directly, but rather as a base class for
    creating chunk table classes with specific vector dimensions.
    """

    uid: PrimaryKeyUUID = msgspec.field(default_factory=PrimaryKeyUUID.factory)
    doc_id: Annotated[UUID, ForeignKey[DefaultDocument.uid]]
    text: str
    vec: Vector[1]  # as a placeholder, will be replaced by the actual vector type
    keyword: Keyword


def create_chunk_with_dim(dim: int) -> Type[_DefaultChunk]:
    """Create a chunk table class with a specific vector dimension.

    This comes with vector and keyword column. It also has a foreign key to the
    :class:`DefaultDocument` table. (If this is used, the :class:`DefaultDocument`
    table must be registered too.)

    Args:
        dim: vector dimension.
    """
    if not isinstance(dim, int) or dim <= 0:
        raise ValueError(f"dim must be a positive integer, not `{type(dim)}({dim})`")

    DenseVector = Vector[dim]

    class DefaultChunk(_DefaultChunk):
        """A chunk table class with a specific vector dimension."""

        vec: DenseVector  # type: ignore

    return DefaultChunk



from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Sequence

import msgspec
import pytrec_eval

from vechord.errors import DecodeStructuredOutputError
from vechord.model import GeminiGenerateRequest, RetrievedChunk, UMBRELAScore
from vechord.provider import GeminiGenerateProvider


class BaseEvaluator(ABC):
    def evaluate(
        self,
        chunk_ids: list[int],
        retrieves: list[list[RetrievedChunk]],
        measures: Sequence[str] = ("map", "ndcg", "recall"),
    ):
        """Evaluate the retrieval results for multiple queries."""
        num = len(chunk_ids)
        qids = list(range(num))
        query_relevance = {
            str(qid): {str(chunk_id): 1}
            for qid, chunk_id in zip(qids, chunk_ids, strict=False)
        }
        evaluator = pytrec_eval.RelevanceEvaluator(
            query_relevance=query_relevance, measures=measures
        )
        res = {
            str(qid): {str(r.uid): 1 / (r.score + 1e-6) for r in retrieve}
            for qid, retrieve in zip(qids, retrieves, strict=False)
        }
        evaluation = evaluator.evaluate(res)
        avg = defaultdict(float)
        for qres in evaluation.values():
            for k, v in qres.items():
                avg[k] = v
        for k in avg:
            avg[k] /= num
        return avg

    @staticmethod
    def evaluate_one(
        truth_id: str,
        resp_ids: list[str],
        measures: Sequence[str] = ("map", "ndcg", "recall"),
    ):
        """Evaluate the retrieval results for a single query."""
        query_relevance = {"0": {str(truth_id): 1}}
        evaluator = pytrec_eval.RelevanceEvaluator(
            query_relevance=query_relevance, measures=measures
        )
        res = {"0": {str(r): 1 / (i + 1) for i, r in enumerate(resp_ids)}}
        evaluation = evaluator.evaluate(res)
        return evaluation["0"]

    @staticmethod
    def calculate_avg_precision(is_relevant: list[bool], total: int) -> float:
        if total == 0 or not is_relevant:
            return 0.0

        precision = []
        relevant = 0
        for i, rel in enumerate(is_relevant, start=1):
            if rel:
                relevant += 1
                precision.append(relevant / i)
        return sum(precision) / len(precision)

    @staticmethod
    def calculate_mrr(is_relevant: list[bool]) -> float:
        for i, rel in enumerate(is_relevant, start=1):
            if rel:
                return 1.0 / i
        return 0.0

    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError


class GeminiEvaluator(BaseEvaluator, GeminiGenerateProvider):
    """Evaluator using Gemini model to generate search queries."""

    def __init__(self, model: str = "gemini-2.5-flash"):
        super().__init__(model)
        self.prompt = """
Given the following chunk of text and the overall document it belongs to, generate 
the most relevant and informative search query that accurately reflects the specific 
information contained within the chunk. Focus on creating a query that someone would 
use to find this exact information within the document. Prioritize using keywords 
and phrases directly from the chunk text. Consider the overall document context to 
refine the query and avoid ambiguity.
"""

    def name(self) -> str:
        return f"gemini_eval_{self.model}"

    async def produce_query(self, doc: str, chunk: str) -> str:
        contents = "\n".join(
            [
                self.prompt,
                f"<chunk> {chunk} </chunk>",
                f"<document> {doc} </document>",
            ]
        )
        resp = await self.query(GeminiGenerateRequest.from_prompt(contents))
        return resp.get_text().strip()


class GeminiUMBRELAEvaluator(BaseEvaluator, GeminiGenerateProvider):
    """Gemini evaluator with the Bing RELevance Assessor (UMBRELA) metric.

    - paper: https://arxiv.org/pdf/2406.06519
    """

    def __init__(self, model: str = "gemini-2.5-flash", relevant_threshold: int = 2):
        super().__init__(model)
        self.relevant_threshold = relevant_threshold
        self.k_values = (3, 5, 10)
        self.score_schema = msgspec.json.schema(UMBRELAScore)
        self.prompt = """
Given a query and a passage, you must provide a score on an
integer scale of 0 to 3 with the following meanings:
0 = represent that the passage has nothing to do with the query,
1 = represents that the passage seems related to the query but
does not answer it,
2 = represents that the passage has some answer for the query,
but the answer may be a bit unclear, or hidden amongst extraneous
information and
3 = represents that the passage is dedicated to the query and
contains the exact answer.
Important Instruction: Assign category 1 if the passage is
somewhat related to the topic but not completely, category 2 if
passage presents something very important related to the entire
topic but also has some extra information and category 3 if the
passage only and entirely refers to the topic. If none of the
above satisfies give it category 0.
Query: {query}
Passage: {passage}
Split this problem into steps:
Consider the underlying intent of the search.
Measure how well the content matches a likely intent of the query (M).
Measure how trustworthy the passage is (T).
Consider the aspects above and the relative importance of each,
and decide on a final score (O). Final score must be an integer.
Do not provide any code in result. Provide each score in the
format of: a single integer without any reasoning.
"""

    def name(self) -> str:
        return f"gemini_umbrela_{self.model}"

    async def estimate(self, query: str, passage: str) -> int:
        content = self.prompt.format(query=query, passage=passage)
        resp = await self.query(
            GeminiGenerateRequest.from_prompt_structure_response(
                content, self.score_schema
            )
        )
        try:
            score = msgspec.json.decode(resp.get_text(), type=UMBRELAScore).score
        except (msgspec.DecodeError, KeyError) as err:
            raise DecodeStructuredOutputError(
                "Failed to decode UMBRELA score from Gemini response",
            ) from err
        return score

    async def evaluate_with_estimation(
        self, query: str, passages: list[str]
    ) -> dict[str, float]:
        """Calculate the Precision@K and Mean Reciprocal Rank (MRR)."""
        scores = [await self.estimate(query, p) for p in passages]
        is_relevant = [score >= self.relevant_threshold for score in scores]
        metric = defaultdict(float)

        for k in self.k_values:
            if k > len(scores) or k <= 0:
                continue
            rel_at_k = sum(is_relevant[:k])
            metric[f"precision@{k}"] = rel_at_k / k
            metric[f"AP@{k}"] = self.calculate_avg_precision(is_relevant[:k], rel_at_k)

        metric["MRR"] = self.calculate_mrr(is_relevant)
        return metric



import asyncio
from os import environ
from typing import Literal

import httpx
import msgspec

from vechord.errors import APIKeyUnsetError, HTTPCallError, TimeoutError
from vechord.model import (
    GeminiEmbeddingRequest,
    GeminiEmbeddingResponse,
    GeminiGenerateRequest,
    GeminiGenerateResponse,
    JinaEmbeddingRequest,
    JinaEmbeddingResponse,
    VoyageEmbeddingRequest,
    VoyageEmbeddingResponse,
    VoyageMultiModalEmbeddingRequest,
)
from vechord.model.llamacloud import LlamaCloudParseRequest, LlamaCloudParseResponse
from vechord.utils import (
    GEMINI_EMBEDDING_RPS,
    GEMINI_GENERATE_RPS,
    JINA_EMBEDDING_RPS,
    VOYAGE_EMBEDDING_RPS,
    RateLimitTransport,
)

EXTRACT_MAX_POLLING_TIME = 1800  # 30 minutes
EXTRACT_CHECK_INTERVAL = 5  # seconds


class BaseProvider:
    PROVIDER_NAME = "UNKNOWN"

    def __init__(self, model: str):
        self.model = model
        api_key_name = f"{self.PROVIDER_NAME.upper()}_API_KEY"
        self.api_key = environ.get(api_key_name)
        if not self.api_key:
            raise APIKeyUnsetError(api_key_name)

    async def __aenter__(self):
        return self

    async def __aexit__(self, _exc_type, _exc_value, _traceback):
        await self.client.aclose()


class GeminiGenerateProvider(BaseProvider):
    """Gemini Generate Provider."""

    PROVIDER_NAME = "GEMINI"

    def __init__(self, model: str = "gemini-2.5-flash"):
        super().__init__(model)
        self.client = httpx.AsyncClient(
            headers={
                "Content-Type": "application/json",
                "x-goog-api-key": self.api_key,
            },
            timeout=httpx.Timeout(120.0, connect=10.0),
            transport=RateLimitTransport(max_per_second=GEMINI_GENERATE_RPS),
        )
        self.encoder = msgspec.json.Encoder()
        self.decoder = msgspec.json.Decoder(GeminiGenerateResponse)
        self.url = (
            "https://generativelanguage.googleapis.com/v1beta/models/"
            f"{self.model}:generateContent"
        )

    async def query(self, req: GeminiGenerateRequest) -> GeminiGenerateResponse:
        """Query the Gemini model with a request."""
        response = await self.client.post(self.url, content=self.encoder.encode(req))
        if response.is_error:
            raise HTTPCallError(
                "Failed to query Gemini generate", response.status_code, response.text
            )
        return self.decoder.decode(response.content)


class GeminiEmbeddingProvider(BaseProvider):
    """Gemini Embedding Provider."""

    PROVIDER_NAME = "GEMINI"

    def __init__(
        self,
        model: str = "gemini-embedding-exp-03-07",
        dim: Literal[768, 1536, 3072] = 3072,
    ):
        super().__init__(model)
        self.dim = dim
        self.client = httpx.AsyncClient(
            headers={
                "Content-Type": "application/json",
                "x-goog-api-key": self.api_key,
            },
            timeout=httpx.Timeout(60.0, connect=10.0),
            transport=RateLimitTransport(max_per_second=GEMINI_EMBEDDING_RPS),
        )
        self.url = (
            "https://generativelanguage.googleapis.com/v1beta/models/"
            f"{self.model}:embedContent"
        )
        self.encoder = msgspec.json.Encoder()
        self.decoder = msgspec.json.Decoder(GeminiEmbeddingResponse)

    async def query(self, req: GeminiEmbeddingRequest) -> GeminiEmbeddingResponse:
        """Query the Gemini embedding model with a request."""
        response = await self.client.post(self.url, content=self.encoder.encode(req))
        if response.is_error:
            raise HTTPCallError(
                "Failed to query Gemini embedding", response.status_code, response.text
            )
        return self.decoder.decode(response.content)


class JinaEmbeddingProvider(BaseProvider):
    """Jina Embedding Provider."""

    PROVIDER_NAME = "JINA"

    def __init__(self, model: str = "jina-embeddings-v4", dim: int = 2048):
        super().__init__(model)
        self.dim = dim
        self.client = httpx.AsyncClient(
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {self.api_key}",
            },
            timeout=httpx.Timeout(60.0, connect=10.0),
            transport=RateLimitTransport(max_per_second=JINA_EMBEDDING_RPS),
        )
        self.url = "https://api.jina.ai/v1/embeddings"
        self.encoder = msgspec.json.Encoder()
        self.decoder = msgspec.json.Decoder(JinaEmbeddingResponse)

    async def query(self, req: JinaEmbeddingRequest) -> JinaEmbeddingResponse:
        """Query the Jina embedding model with a request."""
        response = await self.client.post(self.url, content=self.encoder.encode(req))
        if response.is_error:
            raise HTTPCallError(
                "Failed to query Jina embedding", response.status_code, response.text
            )
        return self.decoder.decode(response.content)


class VoyageEmbeddingProvider(BaseProvider):
    """Voyage Embedding Provider."""

    PROVIDER_NAME = "VOYAGE"

    def __init__(self, model: str = "voyage-3.5", dim: int = 1024):
        super().__init__(model)
        self.dim = dim
        self.client = httpx.AsyncClient(
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {self.api_key}",
            },
            timeout=httpx.Timeout(60.0, connect=10.0),
            transport=RateLimitTransport(max_per_second=VOYAGE_EMBEDDING_RPS),
        )
        self.url = "https://api.voyageai.com/v1/embeddings"
        self.encoder = msgspec.json.Encoder()
        self.decoder = msgspec.json.Decoder(VoyageEmbeddingResponse)

    async def query(
        self, req: VoyageEmbeddingRequest | VoyageMultiModalEmbeddingRequest
    ) -> VoyageEmbeddingResponse:
        """Query the Voyage embedding model with a request."""
        response = await self.client.post(self.url, content=self.encoder.encode(req))
        if response.is_error:
            raise HTTPCallError(
                "Failed to query Voyage embedding", response.status_code, response.text
            )
        return self.decoder.decode(response.content)


class LlamaCloudProvider(BaseProvider):
    """LlamaCloud Provider."""

    PROVIDER_NAME = "LLAMA_CLOUD"

    def __init__(self, model=""):
        super().__init__(model)
        self.client = httpx.AsyncClient(
            headers={
                "Authorization": f"Bearer {self.api_key}",
            },
            timeout=httpx.Timeout(60.0, connect=10.0),
        )
        self.url = "https://api.cloud.llamaindex.ai/api"
        self.encoder = msgspec.json.Encoder()
        self.decoder = msgspec.json.Decoder(LlamaCloudParseResponse)

    async def parse(self, req: LlamaCloudParseRequest) -> LlamaCloudParseResponse:
        files = {"file": (req.filename, req.content, req.mime_type)}
        response = await self.client.post(
            f"{self.url}/parsing/upload",
            files=files,
        )
        if response.is_error:
            raise HTTPCallError(
                "Failed to upload file to LlamaCloud for parsing",
                response.status_code,
                response.text,
            )
        return self.decoder.decode(response.content)

    async def get_text(self, job_id: str) -> str:
        """Get the text result from a LlamaCloud job."""
        loop = asyncio.get_running_loop()
        deadline = loop.time() + EXTRACT_MAX_POLLING_TIME
        while True:
            response = await self.client.get(
                f"{self.url}/parsing/job/{job_id}/result/text"
            )
            if response.is_success:
                return response.json()["text"]
            if loop.time() > deadline:
                raise TimeoutError(
                    f"Polling LlamaCloud job result timed out after {EXTRACT_MAX_POLLING_TIME} seconds."
                )
            await asyncio.sleep(EXTRACT_CHECK_INTERVAL)
